// Copyright (C) 2025 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only

#include "chromadb.h"
#include <QJsonArray>
#include <QJsonDocument>
#include <QJsonObject>
#include <QRestReply>
#include <QCryptographicHash>
#include <QTimer>

using namespace Qt::StringLiterals;


void ChromaDb::sendRequest(
    QString url,
    QJsonObject object,
    std::function<void(QJsonDocument, int)> lambda)
{
    QNetworkRequest request(url);
    request.setRawHeader("Content-Type", "application/json");
    qDebug() << object["ids"] << object["documents"];
    m_restApi.post(request, QJsonDocument(object).toJson(), this, [=](QRestReply &reply) {
        std::optional<QJsonDocument> json = reply.readJson();
        if (reply.isHttpStatusSuccess()) {
            lambda(json ? json.value() : QJsonDocument(), reply.httpStatus());
        } else {
            qDebug() << request.url() << "responded with error:" << reply.errorString();
            setError(true);
        }
    });
}


ChromaDb::ChromaDb(QObject *parent)
    : QObject{parent},
      m_manager(this)
    , m_restApi(&m_manager)
{

}

bool ChromaDb::connected() const
{
    return m_connected;
}

void ChromaDb::connect(bool c)
{
    if (connected()) {
        qDebug() << "Already connected";
    } else {
        sendRequest(m_chromadb_url_base + "collections",
                    QJsonObject({{"name", "embeddings_db"}, {"get_or_create", true}}),
                    [this](auto json, int httpCode) {
                        qDebug() << "Using " << json.object()["id"] << "as chromadb collection";
                        m_chromadb_url_collection = u"collections/%1/"_s.arg(json.object()["id"].toString())
                                                        .toLatin1();
                        m_connected = (httpCode == 200);
                        setError(httpCode != 200);
                        emit connectedChanged();
                    });
    }
}

void ChromaDb::reset()
{
    QUrl url(m_chromadb_url_base + "collections/embeddings_db");
    QNetworkRequest request(url);
    m_restApi.deleteResource(request, this, [=](QRestReply &reply) {
        if (reply.isHttpStatusSuccess()) {
            if (m_connected) {
                m_connected = false;
                connect(true);
            }
        } else {
            qDebug() << url << "responded with error:" << reply.errorString();
        }

    });
}

void ChromaDb::storeEmbeddings(QVariantList embeddings, QVariantList documents)
{
    if (connected()) {
        int i = 0;
        QStringList hashArray;
        for (auto &&doc : documents) {
            hashArray.append(QCryptographicHash::hash(doc.toString().toUtf8(), QCryptographicHash::Md5));
        }
        sendRequest(m_chromadb_url_base + m_chromadb_url_collection + "upsert",
                    QJsonObject({{"embeddings", QJsonArray::fromVariantList(embeddings)},
                                 {"ids", QJsonArray::fromStringList(hashArray)},
                                 {"documents", QJsonArray::fromVariantList(documents)}}),
                    [](auto json, int httpCode) {
                        qDebug() << "Embeddings stored:" << json << httpCode;
                    });
        i++;
    } else if(error()) {
        qDebug() << "ChromaDb not supported";
    } else {
        qDebug() << "Not connected. Store embeddings later";
        QTimer::singleShot(100, this, [=]() { storeEmbeddings(embeddings, documents); });
    }

}

void ChromaDb::fetchEmbeddings(QVariantList hints)
{
    if (connected()) {
        sendRequest(m_chromadb_url_base + m_chromadb_url_collection + "query",
                    QJsonObject({{"query_embeddings", QJsonArray::fromVariantList(hints)}, {"n_results", 3}}),
                    [this](auto json, int httpCode) {
                        qDebug() << json;
                        QVariantList embeddings;
                        if (httpCode == 200)
                            embeddings = json.object()["documents"].toArray().toVariantList();
                        emit embeddingsFound(embeddings);
                    });

    } else if(error()) {
        qDebug() << "ChromaDb not supported";
    } else {
        qDebug() << "Not connected. Fetch embeddings later";
        QTimer::singleShot(100, this, [=]() { fetchEmbeddings(hints); });
    }
}

bool ChromaDb::error() const
{
    return m_error;
}

void ChromaDb::setError(bool newError)
{
    if (m_error == newError)
        return;
    m_error = newError;
    emit errorChanged();
}
