// Copyright (C) 2025 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only

#include "qaimodel.h"
#include "qllmaimodel_p.h"
#include <QJsonArray>
#include <QJsonDocument>
#include <QJsonObject>
#include <QJsonValue>
#include <QNetworkReply>
#include <QNetworkRequest>
#include <QRestReply>
#include <QString>
#include <QFile>

using namespace Qt::StringLiterals;

QLlmAiModel::QLlmAiModel()
    : AiModelPrivateInterface(),
      m_manager(this)
    , m_restApi(&m_manager)
{
    m_chromadb.reset();
    m_chromadb.connect(true);
    connect(this, &QLlmAiModel::embeddingsReceived,
            &m_chromadb, &ChromaDb::storeEmbeddings);
}

static inline void sendRequest(
    QRestAccessManager *restApi,
    QString url,
    QJsonObject object,
    QLlmAiModel *owner,
    std::function<void(QJsonDocument)> lambda)
{
    QNetworkRequest request(url);
    request.setRawHeader("Content-Type", "application/json");
    restApi->post(request, QJsonDocument(object).toJson(), owner, [=](QRestReply &reply) {
        if (!reply.isHttpStatusSuccess()) {
            qDebug() << request.url() << "responded with error" << reply.errorString() << " and status:" << reply.httpStatus();
        } else if (std::optional<QJsonDocument> json = reply.readJson()) {
            lambda(json.value());
        } else {
            qDebug() << "Error. No data received from" << request.url() << reply;
        }
    });
}

void QLlmAiModel::pushData(QVariantList data, int seed)
{
    QString query = m_owner->prompt();
    QJsonArray images;
    for (auto &&i : data) {
        query.append('\n');
        query.append(QString::fromLatin1(i.toByteArray()));

        if (i.canConvert<QUrl>()) {
            QFile file(QUrl(i.toUrl()).toLocalFile());
            if (file.open(QIODevice::ReadOnly) != 0) {
                QByteArray ba = file.readAll();
                QByteArray ba2 = ba.toBase64();
                QString s = QString::fromLatin1(ba2);
                images.append(s);
            }
        }
    }
    qDebug() << this << "[\"prompt\"]: " << query << "[images]" << images.count();

    auto promptResponseReceived = [=](auto json) {
        emit dataReceived(json.object()["response"].toString().toUtf8());
    };

    QPair<QString, QJsonValue> options;
    if (seed != 0) {
        options = {"options", QJsonObject({{"seed", {seed}}})};
    }


    if (m_chromadb.connected()) {
        connect(&m_chromadb, &ChromaDb::embeddingsFound, this, [=](auto embeddings) {
            QString documents;
            for (auto &&i : embeddings) {
                for (auto &&j : i.toList()) {
                    documents.append(j.toString() + '\n');
                }
            }
            auto q = QString("Using this data: {%1}. Respond to this prompt: %2")
                         .arg(documents, query);
            qDebug() << q;
            sendRequest(&m_restApi,
                        m_ollama_url_base + "generate",
                        QJsonObject({{"model", m_owner->model()},
                                     {"prompt", q},
                                     {"stream", false},
                                     options}),
                        this,
                        promptResponseReceived);
        }, Qt::SingleShotConnection);

        sendRequest(&m_restApi,
                    m_ollama_url_base + "embed",
                    QJsonObject({{"model", m_owner->model()},
                                 {"input", query},
                                options}),
                    this,
                    [this](auto json) {
                        m_chromadb.fetchEmbeddings(json.object()["embeddings"].toArray().toVariantList());
                    });
    } else {
        sendRequest(&m_restApi,
                    m_ollama_url_base + "generate",
                    QJsonObject({{"model", m_owner->model()},
                                 {"prompt", query},
                                 {"stream", false},
                                 {"images", images},
                                 options}),
                    this,
                    promptResponseReceived);
    }

}
void QLlmAiModel::setRag(QVariantList documentsToBeStored)
{
    if (documentsToBeStored.isEmpty())
        return;

    auto onEmbeddingsReceived = [=](auto json) {
        auto arrayOfEmbeddings(json.object()["embeddings"].toArray().toVariantList());
        emit embeddingsReceived(arrayOfEmbeddings, documentsToBeStored);
    };

    sendRequest(&m_restApi,
                m_ollama_url_base + "embed",
                QJsonObject({{"model", m_owner->model()},
                             {"input", QJsonArray::fromVariantList(documentsToBeStored)}}),
                this,
                onEmbeddingsReceived);

}
