// Copyright (C) 2025 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only

#include "qtritonmodel_p.h"
#include "qaimodel.h"
#include <QJsonArray>
#include <QJsonDocument>
#include <QJsonObject>
#include <QNetworkReply>
#include <QNetworkRequest>
#include <QRestReply>
#include <QString>
#include <QImage>
#include <QBuffer>
#include <QPixmap>
#include <opencv2/core/version.hpp>
#if CV_MAJOR_VERSION == 2
#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#elif CV_MAJOR_VERSION >= 3
#include <opencv2/core.hpp>
#include <opencv2/highgui.hpp>
#include <opencv2/imgproc.hpp>
#endif
#include <iostream>
#include <fstream>

QTritonModel::QTritonModel()
    : AiModelPrivateInterface(),
      m_manager(this)
    , m_restApi(&m_manager)
{

}


#if CV_MAJOR_VERSION == 4
#define GET_TRANSFORMATION_CODE(x) cv::COLOR_##x
#else
#define GET_TRANSFORMATION_CODE(x) CV_##x
#endif
enum ScaleType { NONE = 0, VGG = 1, INCEPTION = 2 };

int
Preprocess(
    const cv::Mat& img, const std::string& format, int img_type1, int img_type3,
    size_t img_channels, const cv::Size& img_size, const ScaleType scale,
    std::vector<uint8_t>* input_data)
{
  // Image channels are in BGR order. Currently model configuration
  // data doesn't provide any information as to the expected channel
  // orderings (like RGB, BGR). We are going to assume that RGB is the
  // most likely ordering and so change the channels to that ordering.

  cv::Mat sample;
  if ((img.channels() == 3) && (img_channels == 1)) {
    cv::cvtColor(img, sample, GET_TRANSFORMATION_CODE(BGR2GRAY));
  } else if ((img.channels() == 4) && (img_channels == 1)) {
    cv::cvtColor(img, sample, GET_TRANSFORMATION_CODE(BGRA2GRAY));
  } else if ((img.channels() == 3) && (img_channels == 3)) {
    cv::cvtColor(img, sample, GET_TRANSFORMATION_CODE(BGR2RGB));
  } else if ((img.channels() == 4) && (img_channels == 3)) {
    cv::cvtColor(img, sample, GET_TRANSFORMATION_CODE(BGRA2RGB));
  } else if ((img.channels() == 1) && (img_channels == 3)) {
    cv::cvtColor(img, sample, GET_TRANSFORMATION_CODE(GRAY2RGB));
  } else {
    std::cerr << "unexpected number of channels " << img.channels()
              << " in input image, model expects " << img_channels << "."
              << std::endl;
    return -1;
  }

  cv::Mat sample_resized;
  if (sample.size() != img_size) {
    cv::resize(sample, sample_resized, img_size);
  } else {
    sample_resized = sample;
  }

  cv::Mat sample_type;
  sample_resized.convertTo(
      sample_type, (img_channels == 3) ? img_type3 : img_type1);

  cv::Mat sample_final;
  if (scale == ScaleType::INCEPTION) {
    if (img_channels == 1) {
      sample_final = sample_type.mul(cv::Scalar(1 / 127.5));
      sample_final = sample_final - cv::Scalar(1.0);
    } else {
      sample_final =
          sample_type.mul(cv::Scalar(1 / 127.5, 1 / 127.5, 1 / 127.5));
      sample_final = sample_final - cv::Scalar(1.0, 1.0, 1.0);
    }
  } else if (scale == ScaleType::VGG) {
    if (img_channels == 1) {
      sample_final = sample_type - cv::Scalar(128);
    } else {
      sample_final = sample_type - cv::Scalar(123, 117, 104);
    }
  } else {
    sample_final = sample_type;
  }

  // Allocate a buffer to hold all image elements.
  size_t img_byte_size = sample_final.total() * sample_final.elemSize();
  size_t pos = 0;
  input_data->resize(img_byte_size);

  // For NHWC format Mat is already in the correct order but need to
  // handle both cases of data being contiguous or not.
  if (format.compare("FORMAT_NHWC") == 0) {
    if (sample_final.isContinuous()) {
      memcpy(&((*input_data)[0]), sample_final.datastart, img_byte_size);
      pos = img_byte_size;
    } else {
      size_t row_byte_size = sample_final.cols * sample_final.elemSize();
      for (int r = 0; r < sample_final.rows; ++r) {
        memcpy(
            &((*input_data)[pos]), sample_final.ptr<uint8_t>(r), row_byte_size);
        pos += row_byte_size;
      }
    }
  } else {
    // (format.compare("FORMAT_NCHW") == 0)
    //
    // For CHW formats must split out each channel from the matrix and
    // order them as BBBB...GGGG...RRRR. To do this split the channels
    // of the image directly into 'input_data'. The BGR channels are
    // backed by the 'input_data' vector so that ends up with CHW
    // order of the data.
    std::vector<cv::Mat> input_bgr_channels;
    for (size_t i = 0; i < img_channels; ++i) {
      input_bgr_channels.emplace_back(
          img_size.height, img_size.width, img_type1, &((*input_data)[pos]));

      pos += input_bgr_channels.back().total() *
             input_bgr_channels.back().elemSize();
    }

    cv::split(sample_final, input_bgr_channels);
  }

  if (pos != img_byte_size) {
    std::cerr << "unexpected total size of channels " << pos << ", expecting "
              << img_byte_size << std::endl;
    return -1;
  }
  return 0;
}


// KServe (Open Inference Protocol API)
void QTritonModel::pushData(QVariantList data, int seed)
{
    // Load the specified image.
    std::ifstream file(data.first().toByteArray().toStdString());
    std::vector<char> data2;
    file >> std::noskipws;
    std::copy(
        std::istream_iterator<char>(file), std::istream_iterator<char>(),
        std::back_inserter(data2));
    if (data2.empty()) {
      qDebug() << "error: unable to read image file " << data;
      return;
    }

    cv::Mat img = imdecode(cv::Mat(data2), 1);
    const cv::Size cvSize(416, 416);
    std::vector<uint8_t> convertedImage;
    Preprocess(img, "FORMAT_NCHW", CV_32FC1, CV_32FC3, 3, cvSize, INCEPTION, &convertedImage);

    QNetworkRequest request(QUrl("http://localhost:8000/v2/models/" + m_owner->model() + "/infer"));
    //request.setRawHeader("Content-Type", "application/json");
    request.setRawHeader("Content-Type", "application/octet-stream");
    QJsonDocument doc;
    QJsonObject input;
    input["name"] = "input";
    QJsonObject inputParameters;
    //inputParameters["binary_data"] = false;
    inputParameters["binary_data_size"] = static_cast<int>(convertedImage.size());
    input["shape"] = QJsonArray({1, 3, 416, 416});
    input["datatype"] = "FP32";
    input["parameters"] = inputParameters;
    QJsonObject output;
    output["name"] = "confs";
    QJsonObject outputParameters;
    outputParameters["classification"] = 1;
    outputParameters["binary_data"] = false;
    output["parameters"] = outputParameters;
    QJsonArray imageAsJson;

#if 0
    //QByteArrayView view(image.constBits(), image.sizeInBytes());
    for (int i = 0; i < image.sizeInBytes() /3; i++) {
        //imageAsJson.append(((double)view[(i * 3) + 0] * (1 / 127.5)) - 1.0);
        imageAsJson.append(((float)view[(i * 3) + 0] / 255.0));
    }
    qDebug() << imageAsJson.first() << imageAsJson.last();
    for (int i = 0; i < image.sizeInBytes() /3; i++) {
        //imageAsJson.append(((double)view[(i * 3) + 1] * (1.0 / 127.5)) + 1.0);
        imageAsJson.append(((float)view[(i * 3) + 1] / 255.0));
    }
    for (int i = 0; i < image.sizeInBytes() /3; i++) {
        //imageAsJson.append(((double)view[(i * 3) + 2] * (1.0 / 127.5)) + 1.0);
        imageAsJson.append(((float)view[(i * 3) + 2] / 255.0));
    }
    input["data"] = imageAsJson;
#endif

    QJsonObject obj = doc.object();
    obj["id"] = "0";
    obj["model"] = m_owner->model();
    obj["model_version"] = "1";
    obj["inputs"] = QJsonArray({input});
    obj["outputs"] = QJsonArray({output});
    doc.setObject(obj);
    auto docAsJson = doc.toJson(QJsonDocument::Compact);
    request.setRawHeader("Inference-Header-Content-Length", QByteArray::number(docAsJson.size()));
    docAsJson.append(convertedImage);

    m_restApi.post(request, docAsJson, this, [this](QRestReply &reply) {
        QJsonParseError jsonError;
        if (auto json = reply.readJson(&jsonError)) {
            qDebug() << json;
            emit dataReceived(json->toJson(QJsonDocument::Compact));
        }
        else
        {
            qDebug() << reply << jsonError.errorString();
        }
    });
}
