// Copyright (C) 2025 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only

#include "qaimodel.h"
#include "qaimodelinterface_p.h"
#include <QPluginLoader>
#include <QIODevice>
#include <QDir>
#include <QJsonArray>
#include <QList>
#include <QCoreApplication>

QAiModel::QAiModel()
{
}

AiModelPrivateInterface::AiModelTypes QAiModel::type() const
{
    return m_type;
}

static AiModelPrivateInterface::AiModelTypes constructAiModelTypeFlags(QVariantList list)
{
    auto&& metaEnum = QMetaEnum::fromType<AiModelPrivateInterface::AiModelType>();
    AiModelPrivateInterface::AiModelTypes ret;
    for (auto &&flagAsVariant : list) {
        auto string = flagAsVariant.toString().toLatin1();
        auto flag = static_cast<AiModelPrivateInterface::AiModelType>(metaEnum.keysToValue(string));
        ret.setFlag(flag, true);
    }
    return ret;
}

void QAiModel::setType(const AiModelPrivateInterface::AiModelTypes &newType)
{
    qDebug() << "QAiModel::setType(newType: " << newType << ")";

    if (type() == newType)
        return;

    const QObjectList &staticPlugins = QPluginLoader::staticInstances();
    for (auto *plugin : staticPlugins)
        qDebug() << "Static plugin: " << plugin;

    m_interface.clear();
    for (auto &&pluginDir : QCoreApplication::libraryPaths()) {
        QDir aiModelPluginsDir(pluginDir + "/aimodel");
        qDebug() << "Plugins dir: " << aiModelPluginsDir.absolutePath();
        const auto entryList = aiModelPluginsDir.entryList(QDir::Files);
        for (const QString &fileName : entryList) {
            qDebug() << "Loading " << fileName << "...";
            QPluginLoader loader(aiModelPluginsDir.absoluteFilePath(fileName));
            QJsonObject object{ loader.metaData().value("MetaData").toObject() };
            qDebug() << "Metadata for " << fileName << ": " << object;
            if (!object.value("supportedTypes").isArray()) {
                qDebug() << "Incorrect json format in" << loader.metaData()
                         << "for plugin:" << fileName;
                continue;
            }
            auto flagArray = object.value("supportedTypes").toArray().toVariantList();
            auto pluginFlags = constructAiModelTypeFlags(flagArray);

            qDebug() << pluginFlags;
            if (pluginFlags.testFlags(newType)) {
                auto *instance = loader.instance();
                QAiModelPluginFactory *plugin = qobject_cast<QAiModelPluginFactory*>(instance);
                if (plugin) {
                    qDebug() << plugin << "created";
                    m_interface.reset(plugin->createInterface());
                    m_interface->init(this);
                    break;
                } else {
                    qDebug() << "Could not convert" << instance << "to QAiModelPluginFactory*";
                }
            }
    }

    }

    if (!m_interface.isNull()) {
        connect(m_interface.get(),
                &AiModelPrivateInterface::dataReceived,
                this,
                &QAiModel::dataReceived);
        connect(m_interface.get(),
                &AiModelPrivateInterface::processingChanged,
                this,
                &QAiModel::processingChanged);

        m_type = newType;
        emit typeChanged();
    } else {
        qWarning() << "No plugin found for flags:" <<  newType;
    }
}

bool QAiModel::processing() const
{
    return m_processing;
}

void QAiModel::setPrompt(const QString &newPrompt)
{
    if (prompt() == newPrompt)
        return;

    m_prompt = newPrompt;
    emit promptChanged();
}

void QAiModel::setModel(const QString &newModel)
{
    if (model() == newModel)
        return;

    m_model = newModel;
    emit modelChanged();
}

QVector<QAiModel*> QAiModel::inputs() const
{
    return m_inputs;
}

void QAiModel::setInputs(
    QVector<QAiModel*> newInputs)
{
    if (inputs() == newInputs)
        return;
    m_inputs = newInputs;
    if (!m_interface.isNull()) {
        for (auto input: newInputs) {
            input->m_output = this;
        }
        emit inputsChanged();
    }
}

void QAiModel::processCombinedData(QVariant data)
{
    int dataAvailableCounter = 0;
    int optionalInputsWithData = 0;
    QVariantList dataFromAllInputs;

    for (auto &&input : m_inputs) {
        if (!input->m_buffer.isNull()) {
            dataAvailableCounter++;
            dataFromAllInputs.append(input->m_buffer);
        } else if (input->optional() && !input->processing()) {
            qDebug() << "Skipping optional input" << input->interface();
        }
        if (input->processing()) {
            qDebug() << input->interface() << "still processing -> wait for result";
            return;
        }

    }

    pushData(dataFromAllInputs);

    for (auto &&input : m_inputs) {
        m_buffer.clear();
    }
}

int QAiModel::seed() const
{
    return m_seed;
}

void QAiModel::setSeed(int newSeed)
{
    if (m_seed == newSeed)
        return;
    m_seed = newSeed;
    emit seedChanged();
}

void QAiModel::dataReceived(QVariant data)
{
    qDebug() << interface() << ":" << __func__ << "(): data:" << data.typeName();

    m_buffer = data;
    m_processing = false;
    emit processingChanged();
    emit gotResult(data);

    if (m_output) {
        m_output->processCombinedData(data);
    }
}

void QAiModel::pushData(QVariant data)
{
    if (!m_interface.isNull()) {
        if (m_processing) {
            qDebug() << m_interface << "busy. Returning";
            return;
        }
        m_buffer.clear();
        m_processing = true;
        emit processingChanged();
        qDebug() << data.typeName() << data;
        if (QByteArrayView(data.typeName()) == "QVariantList")
            interface()->pushData(data.value<QVariantList>(), m_seed);
        else
            interface()->pushData({data}, m_seed);
    }
}

QVariantList QAiModel::documents() const
{
    return m_documents;
}

void QAiModel::setDocuments(const QVariantList &newDocuments)
{
    if (m_documents == newDocuments)
        return;
    m_documents = newDocuments;
    qDebug() << newDocuments << m_output;
    if (!m_interface.isNull()) {
        m_interface->setRag(newDocuments);
    }

    emit documentsChanged();
}


bool QAiModel::buffered() const
{
    return m_buffered;
}

void QAiModel::setBuffered(bool newBuffered)
{
    if (m_buffered == newBuffered)
        return;
    m_buffered = newBuffered;
    emit bufferedChanged();
}

bool QAiModel::optional() const
{
    return m_optional;
}

void QAiModel::setOptional(bool newOptional)
{
    if (m_optional == newOptional)
        return;
    m_optional = newOptional;
    emit optionalChanged();
}

void QAiModel::clearBuffer()
{
    m_buffer.clear();
}
